from __future__ import annotations

import codecs
import logging
import os
import re
from enum import Enum
from html import entities as htmlentitydefs
from typing import TYPE_CHECKING
from typing import Any
from typing import Generic
from typing import Literal
from typing import NoReturn
from typing import TypeVar
from typing import overload


if TYPE_CHECKING:
    from collections.abc import Callable
    from collections.abc import Iterable
    from collections.abc import Iterator
    from collections.abc import Mapping
    from collections.abc import Sequence
    from types import TracebackType


_KT = TypeVar('_KT')
_VT_co = TypeVar('_VT_co')

log = logging.getLogger('chameleon.utils')


class _Marker(Enum):
    marker = object()


# NOTE: Enums are better markers for type narrowing
marker: Literal[_Marker.marker] = _Marker.marker


def safe_native(s: str | bytes, encoding: str = 'utf-8') -> str:
    if not isinstance(s, str):
        s = bytes.decode(s, encoding, 'replace')
    return s


def raise_with_traceback(
    exc: BaseException,
    tb: TracebackType | None
) -> NoReturn:
    exc.__traceback__ = tb
    raise exc


def encode_string(s: str) -> bytes:
    return bytes(s, 'utf-8')


entity_re = re.compile(r'&(#?)(x?)(\d{1,5}|\w{1,8});')

module_cache = {}

xml_prefixes = (
    (codecs.BOM_UTF8, 'utf-8-sig'),
    (codecs.BOM_UTF16_BE, 'utf-16-be'),
    (codecs.BOM_UTF16_LE, 'utf-16-le'),
    (codecs.BOM_UTF16, 'utf-16'),
    (codecs.BOM_UTF32_BE, 'utf-32-be'),
    (codecs.BOM_UTF32_LE, 'utf-32-le'),
    (codecs.BOM_UTF32, 'utf-32'),
)


def _has_encoding(encoding: str) -> bool:
    try:
        "".encode(encoding)
    except LookupError:
        return False
    else:
        return True


# Precomputed prefix table
_xml_prefixes = tuple(
    (bom, '<?xml'.encode(encoding), encoding)
    for bom, encoding in reversed(xml_prefixes)
    if _has_encoding(encoding)
)

_xml_decl = encode_string("<?xml")

RE_META = re.compile(
    r'\s*<meta\s+http-equiv=["\']?Content-Type["\']?'
    r'\s+content=["\']?([^;]+);\s*charset=([^"\']+)["\']?\s*/?\s*>\s*',
    re.IGNORECASE
)

RE_ENCODING = re.compile(
    br'encoding\s*=\s*(?:"|\')(?P<encoding>[\w\-]+)(?:"|\')',
    re.IGNORECASE
)


def read_encoded(data: bytes) -> str:
    return read_bytes(data, "utf-8")[0]


def read_bytes(
    body: bytes,
    default_encoding: str
) -> tuple[str, str, str | None]:

    for bom, prefix, encoding in _xml_prefixes:
        if body.startswith(bom):
            document = body.decode(encoding)
            return document, encoding, \
                "text/xml" if document.startswith("<?xml") else None

        if prefix != encode_string('<?xml') and body.startswith(prefix):
            return body.decode(encoding), encoding, "text/xml"

    content_type: str | None
    if body.startswith(_xml_decl):
        content_type = "text/xml"
        encoding = read_xml_encoding(body) or default_encoding
    else:
        content_type, encoding = detect_encoding(body, default_encoding)

    return body.decode(encoding), encoding, content_type


def detect_encoding(
    body: str | bytes,
    default_encoding: str
) -> tuple[str | None, str]:

    if not isinstance(body, str):
        body = body.decode('ascii', 'ignore')

    match = RE_META.search(body)
    if match is not None:
        # this can be treated like tuple[str, str] since we unpack it
        return match.groups()  # type: ignore[return-value]

    return None, default_encoding


def read_xml_encoding(body: bytes) -> str | None:
    if body.startswith(b'<?xml'):
        match = RE_ENCODING.search(body)
        if match is not None:
            return match.group('encoding').decode('ascii')
    return None


def mangle(filename: str) -> str:
    """Mangles template filename into top-level Python module name.

    >>> mangle('hello_world.pt')
    'hello_world'

    >>> mangle('foo.bar.baz.pt')
    'foo_bar_baz'

    >>> mangle('foo-bar-baz.pt')
    'foo_bar_baz'

    """

    base, ext = os.path.splitext(filename)
    return base.replace('.', '_').replace('-', '_')


def char2entity(c: str | bytes | bytearray) -> str:
    cp = ord(c)
    name = htmlentitydefs.codepoint2name.get(cp)
    return '&%s;' % name if name is not None else '&#%d;' % cp


def substitute_entity(
    match: re.Match[str],
    n2cp: Mapping[str, int] = htmlentitydefs.name2codepoint
) -> str:
    ent = match.group(3)

    if match.group(1) == "#":
        if match.group(2) == '':
            return chr(int(ent))
        elif match.group(2) == 'x':
            return chr(int('0x' + ent, 16))
        else:
            # FIXME: This should be unreachable, so we can
            #        try raising an AssertionError instead
            return ''
    else:
        cp = n2cp.get(ent)

        if cp:
            return chr(cp)
        else:
            return match.group()


def create_formatted_exception(
    exc: BaseException,
    cls: type[object],
    formatter: Callable[..., str],
    base: type[BaseException] = Exception
) -> BaseException:
    try:
        try:
            new = type(cls.__name__, (cls, base), {
                '__str__': formatter,
                '_original__str__': exc.__str__,
                '__new__': BaseException.__new__,
                '__module__': cls.__module__,
            })
        except TypeError:
            new = cls

        inst: BaseException
        try:
            inst = BaseException.__new__(new)
        except TypeError:
            inst = cls.__new__(new)

        BaseException.__init__(inst, *exc.args)
        inst.__dict__ = exc.__dict__  # type: ignore[attr-defined]

        return inst
    except ValueError:
        name = type(exc).__name__
        log.warn("Unable to copy exception of type '%s'." % name)
        raise TypeError(exc)


_concat = "".join


def join(stream: Iterable[str]) -> str:
    """Concatenate stream.

    >>> print(join(('Hello', ' ', 'world')))
    Hello world

    >>> join(('Hello', 0))
    Traceback (most recent call last):
     ...
    TypeError: ... expected ...

    """

    try:
        return _concat(stream)
    except BaseException:
        # Loop through stream and coerce each element into unicode;
        # this should raise an exception
        for element in stream:
            str(element)

        # In case it didn't, re-raise the original exception
        raise


def decode_htmlentities(string: str) -> str:
    """
    >>> str(decode_htmlentities('&amp;amp;'))
    '&amp;'

    """

    decoded = entity_re.subn(substitute_entity, string)[0]

    # preserve input token data
    return string.replace(string, decoded)


# Taken from zope.dottedname
def _resolve_dotted(name: str, module: str | None = None) -> Any:
    name_parts = name.split('.')
    if not name_parts[0]:
        if module is None:
            raise ValueError("relative name without base module")
        module_parts = module.split('.')
        name_parts.pop(0)
        while not name[0]:
            module_parts.pop()
            name_parts.pop(0)
        name_parts = module_parts + name_parts

    used = name_parts.pop(0)
    found = __import__(used)
    for n in name_parts:
        used += '.' + n
        try:
            found = getattr(found, n)
        except AttributeError:
            __import__(used)
            found = getattr(found, n)

    return found


def resolve_dotted(dotted: str) -> Any:
    if dotted not in module_cache:
        resolved = _resolve_dotted(dotted)
        module_cache[dotted] = resolved
    return module_cache[dotted]


def limit_string(s: str, max_length: int = 53) -> str:
    if len(s) > max_length:
        return s[:max_length - 3] + '...'

    return s


def value_repr(value: object) -> str:
    if isinstance(value, str):
        short = limit_string(value)
        return short.replace('\n', '\\n')
    if isinstance(value, (int, float)):
        return value  # type: ignore[return-value]
    if isinstance(value, dict):
        return '{...} (%d)' % len(value)

    try:
        # FIXME: Is this trailing comma intentional?
        #        it changes the formatting <foo (bar,) at ...>
        #        vs. <foo bar at ...>
        name = str(getattr(value, '__name__', None)),
    except:  # noqa: E722 do not use bare 'except'
        name = '-'  # type: ignore[assignment]

    return '<{} {} at {}>'.format(
        type(value).__name__, name, hex(abs(id(value))))


class callablestr(str):
    __slots__ = ()

    def __call__(self) -> str:
        return self


class callableint(int):
    __slots__ = ()

    def __call__(self) -> int:
        return self


class descriptorstr:
    __slots__ = "function", "__name__"

    def __init__(self, function: Callable[[Any], str]) -> None:
        self.function = function
        self.__name__ = function.__name__

    def __get__(self, context: object, cls: type[object]) -> callablestr:
        return callablestr(self.function(context))


class descriptorint:
    __slots__ = "function", "__name__"

    def __init__(self, function: Callable[[Any], int]) -> None:
        self.function = function
        self.__name__ = function.__name__

    def __get__(self, context: object, cls: type[object]) -> callableint:
        return callableint(self.function(context))


class DebuggingOutputStream(list[str]):
    def append(self, value: str) -> None:
        if not isinstance(value, str):
            raise TypeError(value)
        super().append(value)


class Scope(dict[str, Any]):
    """
    >>> scope = Scope()
    >>> scope['a'] = 1
    >>> copy = scope.copy()

    Setting a local value and then a global value, we expect the local value
    to take precedence.

    >>> copy['a'] = 2
    >>> copy.set_global('a', 3)
    >>> assert copy['a'] == 2

    However, setting a new global value should be immediately visible.

    >>> copy.set_global('b', 1)
    >>> assert copy['b'] == 1

    Make sure the objects are reference-counted, not requiring a full
    collection to be disposed of.

    >>> import gc
    >>> _ = gc.collect()
    >>> del copy
    >>> del scope
    >>> import platform
    >>> assert gc.collect() == (
    ...     0 if platform.python_implementation() == 'CPython' else None
    ... )
    """

    __slots__ = "_root",

    set_local = dict.__setitem__

    def __getitem__(self, key: str) -> Any:
        value = self.get(key, marker)
        if value is marker:
            raise KeyError(key)
        return value

    def __contains__(self, key: object) -> bool:
        return self.get(key, marker) is not marker  # type: ignore

    def __iter__(self) -> Iterator[str]:
        root = getattr(self, "_root", marker)
        yield from super().__iter__()
        if root is not marker:
            for key in root:
                if not super().__contains__(key):
                    yield key

    @overload
    def get(self, key: str, default: None = None) -> Any | None: ...
    @overload
    def get(self, key: str, default: object) -> Any: ...

    def get(self, key: str, default: object = None) -> Any:
        value = super().get(key, marker)
        if value is not marker:
            return value

        root = getattr(self, "_root", marker)
        if root is not marker:
            value = super(Scope, root).get(key, marker)

            if value is not marker:
                return value

        return default

    @property
    def vars(self) -> Mapping[str, Any]:
        return self

    def copy(self) -> Scope:
        inst = Scope(self)
        root = getattr(self, "_root", self)
        inst._root = root  # type: ignore[attr-defined]
        return inst

    def set_global(self, name: str, value: Any) -> None:
        root = getattr(self, "_root", self)
        root[name] = value

    def get_name(self, key: str) -> Any:
        value = self.get(key, marker)
        if value is marker:
            raise NameError(key)
        return value

    setLocal = set_local
    setGlobal = set_global


class ListDictProxy(Generic[_KT, _VT_co]):
    def __init__(
        self: ListDictProxy[_KT, _VT_co],
        _l: Sequence[Mapping[_KT, _VT_co]]
    ) -> None:
        self._l = _l

    def get(self, key: _KT) -> _VT_co | None:
        return self._l[-1].get(key)


class Markup(str):
    """Wraps a string to always render as structure.

    >>> Markup('<br />')
    s'<br />'
    """

    def __html__(self) -> str:
        return str(self)

    def __repr__(self) -> str:
        return "s'%s'" % self


class ImportableMarker:
    def __init__(self, module: str, name: str) -> None:
        self.__module__ = module
        self.name = name

    @property
    def __name__(self) -> str:
        return "%s_MARKER" % self.name

    def __repr__(self) -> str:
        return '<%s>' % self.name


def lookup_attr(obj: object, key: str) -> Any:
    try:
        return getattr(obj, key)
    except AttributeError as exc:
        # FIXME: What are the two try excepts here for?
        #        We just raise the thing we catch...
        try:
            get = obj.__getitem__  # type: ignore[index]
        except AttributeError:
            raise exc
        try:
            return get(key)
        except KeyError:
            raise exc



from __future__ import annotations

from typing import TYPE_CHECKING
from typing import Any
from typing import Literal
from typing import overload

from chameleon.loader import TemplateLoader as BaseLoader
from chameleon.zpt import template


if TYPE_CHECKING:
    from collections.abc import Mapping
    from collections.abc import Sequence
    from typing_extensions import TypeAlias

    _FormatsMapping: TypeAlias = Mapping[str, type[template.PageTemplateFile]]


class TemplateLoader(BaseLoader):
    formats: _FormatsMapping = {
        "xml": template.PageTemplateFile,
        "text": template.PageTextTemplateFile,
    }

    default_format: Literal['xml'] = "xml"

    def __init__(
        self,
        search_path: Sequence[str] | str | None = None,
        default_extension: str | None = None,
        *,
        formats: _FormatsMapping | None = None,
        **kwargs: Any
    ) -> None:

        if formats is not None:
            self.formats = formats

        super().__init__(search_path, default_extension, **kwargs)

    @overload  # type: ignore[override]
    def load(
        self,
        filename: str,
        format: Literal['xml'] | None = None
    ) -> template.PageTemplateFile: ...

    @overload
    def load(
        self,
        filename: str,
        format: Literal['text']
    ) -> template.PageTextTemplateFile: ...

    @overload
    def load(
        self,
        filename: str,
        format: str
    ) -> template.PageTemplateFile: ...

    def load(
        self,
        filename: str,
        format: str | None = None
    ) -> template.PageTemplateFile:
        """Load and return a template file.

        The format parameter determines will parse the file. Valid
        options are `xml` and `text`.
        """

        cls = self.formats[format or self.default_format]
        return super().load(filename, cls)

    __getitem__ = load



from __future__ import annotations

import datetime
import hashlib
import inspect
import logging
import os
import pathlib
import sys
import tempfile
from typing import TYPE_CHECKING
from typing import Any

from chameleon.compiler import Compiler
from chameleon.config import AUTO_RELOAD
from chameleon.config import CACHE_DIRECTORY
from chameleon.config import DEBUG_MODE
from chameleon.config import EAGER_PARSING
from chameleon.exc import ExceptionFormatter
from chameleon.exc import RenderError
from chameleon.exc import TemplateError
from chameleon.loader import MemoryLoader
from chameleon.loader import ModuleLoader
from chameleon.loader import import_package_resource
from chameleon.nodes import Module
from chameleon.utils import DebuggingOutputStream
from chameleon.utils import Scope
from chameleon.utils import create_formatted_exception
from chameleon.utils import detect_encoding
from chameleon.utils import join
from chameleon.utils import mangle
from chameleon.utils import raise_with_traceback
from chameleon.utils import read_bytes
from chameleon.utils import read_xml_encoding
from chameleon.utils import value_repr


PROGRAM_NAME = "initialize"


if TYPE_CHECKING:
    from _typeshed import StrPath
    from abc import abstractmethod
    from collections.abc import Callable
    from collections.abc import Collection

    from chameleon.compiler import ExpressionEngine


if sys.version_info >= (3, 10):
    import importlib.metadata as importlib_metadata
else:
    import importlib_metadata


def safe_get_package_version(name: str) -> str | None:
    try:
        return importlib_metadata.version(name)
    except importlib_metadata.PackageNotFoundError:
        return None


def get_package_versions() -> list[tuple[str, str]]:
    distributions = importlib_metadata.packages_distributions().values()
    versions = {
        x: safe_get_package_version(x) or ""
        for x in sum(distributions, [])
    }
    return sorted(versions.items())


pkg_digest = hashlib.sha1(__name__.encode('utf-8'))
for name, version in get_package_versions():
    pkg_digest.update(name.encode('utf-8'))
    pkg_digest.update(version.encode('utf-8'))


log = logging.getLogger('chameleon.template')


def _make_module_loader() -> ModuleLoader:
    remove = False
    if CACHE_DIRECTORY:
        path = CACHE_DIRECTORY
    else:
        path = tempfile.mkdtemp()
        remove = True

    return ModuleLoader(path, remove)


class BaseTemplate:
    """Template base class.

    Takes a string input which must be one of the following:

    - a string;
    - a utf-8 encoded byte string;
    - a byte string for an XML document that defines an encoding
      in the document premamble;
    - an HTML document that specifies the encoding via the META tag.

    Note that the template input is decoded, parsed and compiled on
    initialization.
    """

    default_encoding = "utf-8"
    default_content_type: str | None = None

    # This attribute is strictly informational in this template class
    # and is used in exception formatting. It may be set on
    # initialization using the optional ``filename`` keyword argument.
    filename: StrPath = '<string>'

    _cooked = False

    loader: ModuleLoader | MemoryLoader
    if DEBUG_MODE or CACHE_DIRECTORY:
        loader = _make_module_loader()
    else:
        loader = MemoryLoader()

    output_stream_factory: type[list]
    if DEBUG_MODE:
        output_stream_factory = DebuggingOutputStream
    else:
        output_stream_factory = list

    debug = DEBUG_MODE

    # The ``builtins`` dictionary can be used by a template class to
    # add symbols which may not be redefined and which are (cheaply)
    # available in the template variable scope
    builtins: dict[str, Any] = {}

    # The ``builtins`` dictionary is updated with this dictionary at
    # cook time. Note that it can be provided at class initialization
    # using the ``extra_builtins`` keyword argument.
    extra_builtins: dict[str, Any] = {}

    # Expression engine must be provided by subclass
    if TYPE_CHECKING:
        @property
        @abstractmethod
        def engine(self) -> Callable[[], ExpressionEngine]: ...
    else:
        engine = None

    # When ``strict`` is set, expressions must be valid at compile
    # time. When not set, this is only required at evaluation time.
    strict = True

    # This should return a value string representation for exception
    # formatting.
    value_repr = staticmethod(value_repr)

    source: str | None

    def __init__(
        self,
        body: str | bytes | None = None,
        **config: Any
    ) -> None:

        self.__dict__.update(config)

        if body is not None:
            self.write(body)

        # This is only necessary if the ``debug`` flag was passed as a
        # keyword argument
        if self.__dict__.get('debug') is True:
            self.loader = _make_module_loader()

    def __call__(self, **kwargs: Any) -> str:
        return self.render(**kwargs)

    def __repr__(self) -> str:
        return "<{} {}>".format(self.__class__.__name__, self.filename)

    if TYPE_CHECKING:
        # since we add config values to our __dict__ they can appear as
        # arbitrary attributes, to clue the type checker into arbitary
        # attribute access being a possibility, we annotate __getattr__
        def __getattr__(self, name: str) -> Any: ...

    @property
    def keep_body(self) -> bool:
        # By default, we only save the template body if we're
        # in debugging mode (to save memory).
        return self.__dict__.get('keep_body', DEBUG_MODE)  # type: ignore

    @property
    def keep_source(self) -> bool:
        # By default, we only save the generated source code if we're
        # in debugging mode (to save memory).
        return self.__dict__.get('keep_source', DEBUG_MODE)  # type: ignore

    def cook(self, body: str) -> None:
        builtins_dict = self.builtins.copy()
        builtins_dict.update(self.extra_builtins)
        names, builtins = zip(*sorted(builtins_dict.items()))
        digest = self.digest(body, names)
        program = self._cook(body, digest, names)

        init = program[PROGRAM_NAME]
        functions = init(*builtins)

        for name, function in functions.items():
            setattr(self, "_" + name, function)

        self._cooked = True

        if self.keep_body:
            self.body = body

    def cook_check(self) -> bool | None:
        assert self._cooked
        return None

    def parse(self, body: str) -> Any:
        raise NotImplementedError("Must be implemented by subclass.")

    def render(self, **__kw: Any) -> str:
        econtext = Scope(__kw)
        rcontext: dict[str, Any] = {}
        self.cook_check()
        stream = self.output_stream_factory()
        target_language = __kw.get("target_language")
        try:
            self._render(
                stream,
                econtext,
                rcontext,
                target_language=target_language
            )
        except RecursionError:
            raise
        except BaseException:
            cls, exc, tb = sys.exc_info()
            try:
                errors = rcontext.get('__error__')
                if errors:
                    formatter = exc.__str__
                    if isinstance(formatter, ExceptionFormatter):
                        if errors is not formatter._errors:
                            formatter._errors.extend(errors)
                        raise

                    formatter = ExceptionFormatter(
                        errors, econtext, rcontext, self.value_repr)

                    try:
                        exc = create_formatted_exception(
                            exc, cls, formatter, RenderError  # type: ignore
                        )
                    except TypeError:
                        pass

                    assert exc is not None
                    raise_with_traceback(exc, tb)

                raise
            finally:
                del exc, tb

        return join(stream)

    def write(self, body: str | bytes) -> None:
        encoding: str | None
        if isinstance(body, bytes):
            body, encoding, content_type = read_bytes(
                body, self.default_encoding
            )
        elif body.startswith('<?xml'):
            content_type = 'text/xml'
            encoding = read_xml_encoding(body.encode("utf-8"))
        else:
            content_type, encoding = detect_encoding(
                body, self.default_encoding
            )

        self.content_type = content_type or self.default_content_type
        self.content_encoding = encoding

        self.cook(body)

    def _get_module_name(self, name: str) -> str:
        return "%s.py" % name

    def _cook(
        self,
        body: str,
        name: str,
        builtins: Collection[str]
    ) -> dict[str, Any]:

        filename = self._get_module_name(name)
        if DEBUG_MODE or (cooked := self.loader.get(filename)) is None:
            try:
                source = self._compile(body, builtins)
                if self.debug:
                    source = "# template: {}\n#\n{}".format(
                        self.filename, source)
                if self.keep_source:
                    self.source = source
                cooked = self.loader.build(source, filename)
            except TemplateError as exc:
                # normalize to str
                exc.token.filename = str(self.filename)
                raise
        elif self.keep_source:
            module_name = cooked.get('__name__')
            module = sys.modules.get(module_name) if module_name else None
            if module is not None:
                self.source = inspect.getsource(module)
            else:
                self.source = None
        return cooked

    def digest(self, body: str, names: Collection[str]) -> str:
        class_name = type(self).__name__.encode('utf-8')
        sha = pkg_digest.copy()
        sha.update(body.encode('utf-8', 'ignore'))
        sha.update(class_name)
        digest = sha.hexdigest()

        filename = str(self.filename)
        if filename and filename != BaseTemplate.filename:
            digest = os.path.splitext(filename)[0] + '-' + digest

        return digest

    def _compile(self, body: str, builtins: Collection[str]) -> str:
        program = self.parse(body)
        module = Module(PROGRAM_NAME, program)
        compiler = Compiler(
            self.engine,
            module,
            str(self.filename),
            body,
            builtins=builtins,
            strict=self.strict
        )
        return compiler.code  # type: ignore[no-any-return]


class BaseTemplateFile(BaseTemplate):
    """File-based template base class.

    Relative path names are supported only when a template loader is
    provided as the ``loader`` parameter.
    """

    # Auto reload is not enabled by default because it's a significant
    # performance hit
    auto_reload = AUTO_RELOAD

    _v_last_read: float | None

    def __init__(
        self,
        filename: StrPath,
        auto_reload: bool | None = None,
        package_name: str | None = None,
        post_init_hook: Callable[[], object] | None = None,
        **config: Any,
    ) -> None:

        if package_name is None:
            # Normalize filename
            filename = os.path.abspath(
                os.path.normpath(os.path.expanduser(filename))
            )
        self.package_name = package_name
        self.filename = filename

        # Override reload setting only if value is provided explicitly
        if auto_reload is not None:
            self.auto_reload = auto_reload

        super().__init__(**config)

        if post_init_hook is not None:
            post_init_hook()

        if EAGER_PARSING:
            self.cook_check()

    def cook_check(self) -> bool:
        if self.auto_reload:
            mtime = self.mtime()

            if mtime != self._v_last_read:
                self._v_last_read = mtime
                self._cooked = False

        if self._cooked is False:
            body = self.read()
            log.debug("cooking %r (%d bytes)..." % (self.filename, len(body)))
            self.cook(body)
            return True

        return False

    def mtime(self) -> float:
        filename = self.filename
        if self.package_name is not None:
            with import_package_resource(self.package_name) as path:
                joined = path.joinpath(filename)
                if isinstance(joined, pathlib.Path):
                    return joined.stat().st_mtime

                # Assume this is a zip-like path object:
                filename = joined.at
                timetuple = path.root.getinfo(filename).date_time
                return datetime.datetime(*timetuple).timestamp()
        try:
            return os.path.getmtime(filename)
        except OSError:
            return 0

    def read(self) -> str:
        if self.package_name is not None:
            with import_package_resource(self.package_name) as files:
                path = files.joinpath(self.filename)
                data = path.read_bytes()
        else:
            with open(self.filename, "rb") as f:
                data = f.read()

        body, encoding, content_type = read_bytes(data, self.default_encoding)

        self.content_type = content_type or self.default_content_type
        self.content_encoding = encoding

        return body

    def _get_module_name(self, name: str) -> str:
        filename = os.path.basename(str(self.filename))
        mangled = mangle(filename)
        return "{}_{}.py".format(mangled, name)

    def _get_filename(self) -> StrPath:
        # FIXME: Shouldn't a missing filename be an error?
        return self.__dict__.get('filename')  # type: ignore[return-value]

    def _set_filename(self, filename: StrPath) -> None:
        self.__dict__['filename'] = filename
        self._v_last_read = None
        self._cooked = False

    if TYPE_CHECKING:
        # type checkers only understand this way of defining properties
        @property
        def filename(self) -> StrPath: ...
        @filename.setter
        def filename(self, filename: StrPath) -> None: ...
    else:
        filename = property(_get_filename, _set_filename)



# flake8: noqa: F401 unused

class TestImports:
    def test_pagetemplates(self):
        from chameleon import PageTemplate
        from chameleon import PageTemplateFile
        from chameleon import PageTemplateLoader

    def test_pagetexttemplates(self):
        from chameleon import PageTextTemplate
        from chameleon import PageTextTemplateFile

    def test_exceptions(self):
        from chameleon.exc import ExpressionError

    def test_compiler_utils(self):
        from chameleon.astutil import Builtin
        from chameleon.astutil import NameLookupRewriteVisitor
        from chameleon.astutil import Symbol
        from chameleon.codegen import template
        from chameleon.compiler import ExpressionEngine
        from chameleon.compiler import ExpressionEvaluator
        from chameleon.tales import ExpressionParser
        from chameleon.utils import Scope



import difflib
import unittest

from chameleon import PageTemplate


xml_bytes = b"""\
<?xml version="1.0" ?>
<input type="checkbox" checked="nope" tal:attributes="checked checked" />
<input type="checkbox" checked="${checked}" />
"""

xml_w_enc_bytes = b"""\
<?xml version="1.0" encoding="ascii" ?>
<input type="checkbox" checked="nope" tal:attributes="checked checked" />
<input type="checkbox" checked="${checked}" />
"""

html5_bytes = b"""\
<!DOCTYPE html>
<html>
<head>
  <title>Title of document</title>
</head>
<body>
  <form>
    <input type="checkbox" checked="nope"
           tal:attributes="checked checked" />
    <input type="checkbox" checked="${checked}" />
  </form>
</body>
</html>
"""

html5_w_ct_n_enc_bytes = b"""\
<!DOCTYPE html>
<html>
<head>
  <meta http-equiv="Content-Type" content="foo/bar; charset=utf-8" />
  <title>Title of document</title>
</head>
<body>
  <form>
    <input type="checkbox" checked="nope"
           tal:attributes="checked checked" />
    <input type="checkbox" checked="${checked}" />
  </form>
</body>
</html>
"""


class BaseTestCase(unittest.TestCase):
    def get_template(self, text):
        template = PageTemplate(text)
        return template

    def get_template_bytes(self):
        return self.get_template(self.input_bytes)

    def get_template_str(self):
        return self.get_template(self.input_bytes.decode('utf-8'))

    def assert_same(self, s1, s2):
        L1 = s1.splitlines()
        L1 = list(filter(None, [' '.join(x.split()).strip() for x in L1]))
        L2 = s2.splitlines()
        L2 = list(filter(None, [' '.join(x.split()).strip() for x in L2]))
        diff = '\n'.join(list(difflib.unified_diff(L1, L2)))
        assert diff == '', diff


class XMLTestCase(BaseTestCase):

    input_bytes = xml_bytes
    encoding = None

    def test_bytes_content_type(self):
        template = self.get_template_bytes()
        self.assertEqual(template.content_type, 'text/xml')

    def test_bytes_encoding(self):
        template = self.get_template_bytes()
        self.assertEqual(template.content_encoding, 'utf-8')

    def test_str_content_type(self):
        template = self.get_template_str()
        self.assertEqual(template.content_type, 'text/xml')

    def test_str_encoding(self):
        template = self.get_template_str()
        self.assertEqual(template.content_encoding, self.encoding)

    def test_bytes_checked_true(self):
        template = self.get_template_bytes()
        expected = """
        <?xml version="1.0" ?>
        <input type="checkbox" checked="True" />
        <input type="checkbox" checked="True" />
        """
        result = template(checked=True)
        self.assert_same(expected, result)

    def test_bytes_checked_false(self):
        template = self.get_template_bytes()
        expected = """
        <?xml version="1.0" ?>
        <input type="checkbox" checked="False" />
        <input type="checkbox" checked="False" />
        """
        result = template(checked=False)
        self.assert_same(expected, result)

    def test_bytes_checked_None(self):
        template = self.get_template_bytes()
        expected = """
        <?xml version="1.0" ?>
        <input type="checkbox" />
        <input type="checkbox" />
        """
        result = template(checked=None)
        self.assert_same(expected, result)

    def test_bytes_checked_default(self):
        template = self.get_template_bytes()
        expected = """
        <?xml version="1.0" ?>
        <input type="checkbox" checked="nope" />
        <input type="checkbox" />
        """
        result = template(checked=template.default_marker.value)
        self.assert_same(expected, result)

    def test_str_checked_true(self):
        template = self.get_template_str()
        expected = """
        <?xml version="1.0" ?>
        <input type="checkbox" checked="True" />
        <input type="checkbox" checked="True" />
        """
        result = template(checked=True)
        self.assert_same(expected, result)

    def test_str_checked_false(self):
        template = self.get_template_str()
        expected = """
        <?xml version="1.0" ?>
        <input type="checkbox" checked="False" />
        <input type="checkbox" checked="False" />
        """
        result = template(checked=False)
        self.assert_same(expected, result)

    def test_str_checked_None(self):
        template = self.get_template_str()
        expected = """
        <?xml version="1.0" ?>
        <input type="checkbox" />
        <input type="checkbox" />
        """
        result = template(checked=None)
        self.assert_same(expected, result)

    def test_str_checked_default(self):
        template = self.get_template_str()
        expected = """
        <?xml version="1.0" ?>
        <input type="checkbox" checked="nope" />
        <input type="checkbox" />
        """
        result = template(checked=template.default_marker.value)
        self.assert_same(expected, result)


class XMLWithEncodingTestCase(BaseTestCase):

    input_bytes = xml_w_enc_bytes
    encoding = 'ascii'

    def test_bytes_encoding(self):
        template = self.get_template_bytes()
        self.assertEqual(template.content_encoding, self.encoding)

    def test_str_encoding(self):
        template = self.get_template_str()
        self.assertEqual(template.content_encoding, self.encoding)


class HTML5TestCase(BaseTestCase):

    input_bytes = html5_bytes

    def test_bytes_content_type(self):
        template = self.get_template_bytes()
        self.assertEqual(template.content_type, 'text/html')

    def test_bytes_encoding(self):
        template = self.get_template_bytes()
        self.assertEqual(template.content_encoding, 'utf-8')

    def test_str_content_type(self):
        template = self.get_template_str()
        self.assertEqual(template.content_type, 'text/html')

    def test_str_encoding(self):
        template = self.get_template_str()
        self.assertEqual(template.content_encoding, 'utf-8')

    def test_bytes_checked_true(self):
        template = self.get_template_bytes()
        expected = """
        <!DOCTYPE html>
        <html>
        <head>
          <title>Title of document</title>
        </head>
        <body>
          <form>
            <input type="checkbox" checked="checked" />
            <input type="checkbox" checked="checked" />
          </form>
        </body>
        </html>
        """
        result = template(checked=True)
        self.assert_same(expected, result)

    def test_bytes_checked_false(self):
        template = self.get_template_bytes()
        expected = """
        <!DOCTYPE html>
        <html>
        <head>
          <title>Title of document</title>
        </head>
        <body>
          <form>
            <input type="checkbox" />
            <input type="checkbox" />
          </form>
        </body>
        </html>
        """
        result = template(checked=False)
        self.assert_same(expected, result)

    def test_bytes_checked_None(self):
        template = self.get_template_bytes()
        expected = """
        <!DOCTYPE html>
        <html>
        <head>
          <title>Title of document</title>
        </head>
        <body>
          <form>
            <input type="checkbox" />
            <input type="checkbox" />
          </form>
        </body>
        </html>
        """
        result = template(checked=None)
        self.assert_same(expected, result)

    def test_bytes_checked_default(self):
        template = self.get_template_bytes()
        expected = """
        <!DOCTYPE html>
        <html>
        <head>
          <title>Title of document</title>
        </head>
        <body>
          <form>
            <input type="checkbox" checked="nope" />
            <input type="checkbox" />
          </form>
        </body>
        </html>
        """
        result = template(checked=template.default_marker.value)
        self.assert_same(expected, result)

    def test_str_checked_true(self):
        template = self.get_template_str()
        expected = """
        <!DOCTYPE html>
        <html>
        <head>
          <title>Title of document</title>
        </head>
        <body>
          <form>
            <input type="checkbox" checked="checked" />
            <input type="checkbox" checked="checked" />
          </form>
        </body>
        </html>
        """
        result = template(checked=True)
        self.assert_same(expected, result)

    def test_str_checked_false(self):
        template = self.get_template_str()
        expected = """
        <!DOCTYPE html>
        <html>
        <head>
          <title>Title of document</title>
        </head>
        <body>
          <form>
            <input type="checkbox" />
            <input type="checkbox" />
          </form>
        </body>
        </html>
        """
        result = template(checked=False)
        self.assert_same(expected, result)

    def test_str_checked_None(self):
        template = self.get_template_str()
        expected = """
        <!DOCTYPE html>
        <html>
        <head>
          <title>Title of document</title>
        </head>
        <body>
          <form>
            <input type="checkbox" />
            <input type="checkbox" />
          </form>
        </body>
        </html>
        """
        result = template(checked=None)
        self.assert_same(expected, result)

    def test_str_checked_default(self):
        template = self.get_template_str()
        expected = """
        <!DOCTYPE html>
        <html>
        <head>
          <title>Title of document</title>
        </head>
        <body>
          <form>
            <input type="checkbox" checked="nope" />
            <input type="checkbox" />
          </form>
        </body>
        </html>
        """
        result = template(checked=template.default_marker.value)
        self.assert_same(expected, result)


class HTML5WithContentTypeAndEncodingTestCase(BaseTestCase):

    input_bytes = html5_w_ct_n_enc_bytes

    def test_bytes_content_type(self):
        template = self.get_template_bytes()
        self.assertEqual(template.content_type, 'foo/bar')

    def test_bytes_encoding(self):
        template = self.get_template_bytes()
        self.assertEqual(template.content_encoding, 'utf-8')

    def test_str_content_type(self):
        template = self.get_template_str()
        self.assertEqual(template.content_type, 'foo/bar')

    def test_str_encoding(self):
        template = self.get_template_str()
        self.assertEqual(template.content_encoding, 'utf-8')



import unittest


class ScopeTestCase(unittest.TestCase):
    def setUp(self):
        from chameleon.utils import Scope
        scope = Scope()
        scope['a'] = 1
        scope.set_global("b", 2)

        self.parent = scope
        self.child = scope.copy()

    def test_items(self):
        assert list(self.child.items()) == [('a', 1), ('b', 2)]

    def test_keys(self):
        assert list(self.child.keys()) == ['a', 'b']

    def test_values(self):
        assert list(self.child.values()) == [1, 2]



from chameleon.exc import TemplateError
from chameleon.zpt.loader import TemplateLoader as PageTemplateLoader
from chameleon.zpt.template import PageTemplate
from chameleon.zpt.template import PageTemplateFile
from chameleon.zpt.template import PageTextTemplate
from chameleon.zpt.template import PageTextTemplateFile


__all__ = (
    'TemplateError',
    'PageTemplateLoader',
    'PageTemplate',
    'PageTemplateFile',
    'PageTextTemplate',
    'PageTextTemplateFile',
)



__version__ = '4.5.3'

import os

from setuptools import find_packages
from setuptools import setup
from setuptools.command.test import test


here = os.path.abspath(os.path.dirname(__file__))
try:
    README = open(os.path.join(here, 'README.rst')).read()
    CHANGES = open(os.path.join(here, 'CHANGES.rst')).read()
except BaseException:  # doesn't work under tox/pip
    README = ''
    CHANGES = ''

install_requires = ['importlib-metadata;python_version<"3.10"']


class Benchmark(test):
    description = "Run benchmarks"

    def finalize_options(self):
        self.distribution.tests_require = [
            'zope.pagetemplate',
            'zope.component',
            'zope.i18n',
            'zope.testing']

        test.finalize_options(self)

    def run_tests(self):
        from chameleon import benchmark
        print("running benchmark...")

        benchmark.start()


setup(
    name="Chameleon",
    version=__version__,
    description="Fast HTML/XML Template Compiler.",
    long_description="\n\n".join((README, CHANGES)),
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3.13",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    author="Malthe Borch",
    author_email="mborch@gmail.com",
    url="https://chameleon.readthedocs.io",
    project_urls={
        'Documentation': 'https://chameleon.readthedocs.io',
        'Issue Tracker': 'https://github.com/malthe/chameleon/issues',
        'Sources': 'https://github.com/malthe/chameleon',
    },
    license='BSD-like (http://repoze.org/license.html)',
    packages=find_packages('src'),
    package_dir={'': 'src'},
    include_package_data=True,
    package_data={
        'chameleon': [
            'py.typed',
        ],
    },
    python_requires='>=3.9',
    install_requires=install_requires,
    extras_require={
        'docs': {
            'Sphinx',
            'sphinx_rtd_theme',
        },
    },
    zip_safe=False,
    cmdclass={
        'benchmark': Benchmark,
    }
)
